# Copyright (c) 2007, David Bender, Benegon Enterprises LLC
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

#   * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
#   * Neither the name of the Benegon Enterprises LLC nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The diskegrator is hosted at
# http://diskegrator.sourceforge.net

import sys
import subprocess
import os

def zeroFile(path, size):
	count = (size + 32767) /32768 
	print size, count
	cmd = ['dd', 'if=/dev/zero', 'of='+path, 'bs=32k', 'count='+str(count)]
	retval = subprocess.Popen(cmd).wait()
	if retval != 0:
		raise IOError('Unable to zero file: ' + str(path))

def readPartGeometry(devNode):
	# Attempt to read geometry stored in block device
	cmd = ['sfdisk', '-G', devNode]
	proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
	line = proc.stdout.read()
	print "sfdisk output: ",line
	retval = proc.wait()
	if retval != 0:
		raise Exception, 'Unable to infer geometry of ' + str(devNode)
	# parse geometry

	words = line.split()
	return (int(words[1]), int(words[3]), int(words[5]))

# Write partition table to the block device at devNode
# Return a list of (offset, length) of each partition, in byte counts
def writePartitionTable(devNode, geom, partitions):
	locations = []
	offset = 63
	# Generate sfdisk commands
	# Create list of offsets as well
	sfdisk = ""
	partCount = 0
	for entry in partitions:
		partCount += 1
		numSectors = (entry.length() + 511) / 512
		sfdisk += str(offset)+','+str(numSectors)
		# FIXME primary, extended, logical, blah
		sfdisk += '\n'
		locations.append((offset * 512, numSectors * 512))
		offset += numSectors
		pass
	
	# Unused primary partitions must be set to 0,0
	while partCount < 4:
		sfdisk += '0,0\n'
		partCount += 1

	# Execute sfdisk, ignore warnings from sfdisk
	cmd = ['sfdisk', '-uS', '-L', '--force', devNode]
	cmd.extend(['-C', str(geom[0]), '-H', str(geom[1]), '-S', str(geom[2])])
	proc = subprocess.Popen(cmd, stdin=subprocess.PIPE)
	proc.communicate(sfdisk)
	retval = proc.wait()
	if retval != 0:
		raise Exception, 'Unable to write partition table'

	# Return offsets calculated
	return locations

# Return a list of (offset, length) of each partition
def readPartitionTable(devNode):
	cmd = ['sfdisk', '-luS', '--force', devNode]
	proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
	locations = []

	# Each line is of the form
	#  Device Boot    Start       End   #sectors  Id  System
	# /dev/sda1            63   4016249    4016187  83  Linux
	# 0                    1    2          3   4      5        6
	# devNode : start=       63, size=  4016187, Id=83
	for line in proc.stdout:
		# Tokenize into words
		print "sfdisk: ", line
		words = line.split()
		if not len(words):
			continue

		if words[0].startswith(devNode) and words[5] != "Empty":
			# get start and length, remove comma
			start = int(words[1]) * 512
			size = int(words[3]) * 512
			locations.append((start, size))
	ret = proc.wait()

	if ret != 0:
		raise Exception, 'Unable to read partition table'
	return locations

def sumPartitions(parts):
	total = 63 * 512
	for entry in parts:
		fsize = entry.length()
		fsize += 511
		fsize &= 0xFFFFFFFFFE00
		total += fsize
	return total

class Target:
	# ctor
	def __init__(self):
		pass
	
	# Calling this method will guarantee the availabilty of device nodes for the device and its partitions
	def initialize(self, partitions):
		pass
	
	# Retrieve block device given the index
	# Note that index of 0 will return device's node
	def getDevNode(self, index):
		pass
	
	# Return (cylinders, heads, sectors)
	def geometry(self):
		return self.geometry


class BlockDevice(Target):
	# ctor
	def __init__(self, devNode, geo=None):
		Target.__init__(self)
		self.devNode = devNode
		if geo:
			self.geometry = geo
		else:
			self.geometry = readPartGeometry(self.devNode)

	def getDevNode(self, index):
		retval = self.devNode
		if index > 0:
			retval += str(index)
		return retval

	def initialize(self, partitions):
		self.exists = True

		# Perform the write as a way to allocate space...
		fsize = sumPartitions(partitions)

		# FIXME Ensure that disk size is sufficient

		# Initialize partition table
		locations = writePartitionTable(self.devNode, self.geometry, partitions)

class Loopback(Target):
	def __init__(self, file, fakeDev='/dev/fakedevice'):
		Target.__init__(self)

		# initialize defaults
		self.filename = file
		self.loopIndex = {}
		self.loopCount = 0
		self.fakeDev = fakeDev

		# If file exists, then directly initialize from the file
		try:
			os.stat(file)
			if not os.access(file, os.R_OK | os.W_OK):
				raise IOError(file + " is not accessible!")

			self.addLoopDev()
			locations = readPartitionTable(self.fakeDev)

			# Create the devnodes for each partition
			for (offset, length) in locations:
				self.addLoopDev(['-o', str(offset), '--sizelimit', str(length)])

			self.exists = True 

		except:
			self.exists = False



	def __del__(self):
		# Remove symbolic links to device nodes
		# Free device nodes
		for x in range(1, len(self.loopIndex)):
			cmd = ['losetup', '-d', self.loopIndex[x]]
			retval = subprocess.Popen(cmd).wait()
			if retval != 0:
				print 'Warning: Unable to destroy ' + self.loopIndex[x]
			os.remove(self.fakeDev+str(x))

		# Free root device node
		cmd = ['losetup', '-d', self.loopIndex[0]]
		retval = subprocess.Popen(cmd).wait()
		if retval != 0:
			print 'Warning: Unable to unsetup ', self.loopIndex[0]
		os.remove(self.fakeDev)

	def initialize(self, partitions):
		if self.loopCount < 2:
			# Perform the write as a way to allocate space...
			fsize = sumPartitions(partitions)
			zeroFile(self.filename, fsize)
			self.exists = True

			self.geometry = (fsize / 512 / 63 / 255, 255, 63)
			print "Geometry ", self.geometry

			# Initialize partition table
			locations = writePartitionTable(self.fakeDev, self.geometry, partitions)
			print locations

			# Create the devnodes for each partition
			for (offset, length) in locations:
				self.addLoopDev(['-o', str(offset), '--sizelimit', str(length)])

	def getDevNode(self, index):
		# Check if we have mounted this yet
		if not self.loopIndex.has_key(index):
			raise Exception, 'Partition ' + str(index) + ' not found!'
		retval = self.fakeDev
		if index > 0:
			retval += str(index)
		return retval
	
	def addLoopDev(self, opts = None):
		while self.loopCount < 8:
			loopDev = '/dev/loop' + str(self.loopCount)
			cmd = ['losetup', loopDev, self.filename]
			if opts:
				cmd.extend(opts)
			retval = subprocess.Popen(cmd).wait()
			if retval == 0:
				# Success, increment device count
				self.loopCount += 1

				# Create fake device for it
				target = self.fakeDev
				numParts = len(self.loopIndex)
				if numParts > 0:
					target += str(numParts)
				os.symlink(loopDev, target)
				self.loopIndex[numParts] = loopDev
				return target
		raise 'Unable to map to loop device!'
